#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Set', '_BaseSet', '_SetContainer', '_SetArray', '_ProductSet',
           'set_options']
import logging
import sys
import types
import copy
import pyutilib.misc
from pyutilib.component.core import alias
from component import *
import itertools

try:
    import bidict
    using_bidict=True
except:
    using_bidict=False

logger = logging.getLogger('coopr.pyomo')

#
# Decorator for set initializer functions
#
def set_options(**kwds):
    def decorator(func):
        func.set_options = kwds
        return func
    return decorator


##------------------------------------------------------------------------
##
## An abstract Set class
##
##------------------------------------------------------------------------

class _BaseSet(Component):
    """The base class for set objects that are used to index other Pyomo objects

       This class has a similar look-and-feel as a built-in set class.  However,
       the set operations defined in this class return another abstract
       Set object.  This class contains a concrete set, which can be 
       initialized by the load() method.
    """

    """ Constructor
        Arguments:
           name          The name of the set
           within        A set that defines the type of values that can
                          be contained in this set
           rule
           initialize    Default set members, which may be overriden
                          when setting up this set
           validate      Define a function for validating membership in a
                          set.  This has the functional form:
                          f: data -> bool
                          and returns true if the data belongs in the set
           dimen         Specify the set's arity.
           doc           Documentation for this object
           virtual       Indicate that this is a virtual set that does not
                            have concrete members.
    """
    def __init__(self, **kwds):
        Component.__init__(self, ctype=Set)

        self.order=[]
        # IMPT: order_dict keys are 1-based.
        if using_bidict:
            self.order_dict = bidict.bidict()
        else:
            self.order_dict = {}
            self.order_dict_inv = {}

        # 'rule' and 'initialize' are mutually exclusive options
        if "rule" in kwds and "initialize" in kwds:
            raise TypeError, "Cannot specify both 'rule' and 'initialize' " + \
                  "keywords in '%s'" % str(self)

        # Get keyword arguments
        self.initialize = kwds.pop("rule", None)
        self.initialize = kwds.pop("initialize", self.initialize)
        self.name       = kwds.pop("name", "_unknown_")
        self.validate   = kwds.pop("validate", None)
        self.ordered    = kwds.pop("ordered", False)
        self.virtual    = kwds.pop("virtual", False)
        self.doc        = kwds.pop("doc", "")
        self._bounds    = kwds.pop("bounds", None)
        self.filter     = kwds.pop("filter", None)

        # We can't access self.dimen after its been written, so we use
        # tmp_dimen until the end of __init__
        tmp_dimen = None

        # Get dimen from within, if possible
        self.domain = kwds.pop("within", None)
        if self.domain is not None:
            tmp_dimen = self.domain.dimen

        # Make sure dimen and implied dimensions don't conflict
        kwd_dimen = kwds.pop("dimen", None)
        if kwd_dimen is not None:
            if self.within is not None and tmp_dimen != kwd_dimen:
                raise ValueError, \
                      ("Value of keyword 'dimen', %s, differs from the " + \
                       "dimension of the superset '%s', %s") % \
                       (str(kwd_dimen), str(self.within.name), str(tmp_dimen))
            else:
                tmp_dimen = kwd_dimen

        # See there were any erroneous keyword arguments
        if len(kwds) != 0:
            msg = "Unknown keyword arguments in constructor of '%s': " % \
                  str(self)

            # Add unknown keyword arguments
            for (k, v) in kwds.items():
                msg += "%s=%s, " % (k, v)

            # Clip trailing ', '
            msg = msg[:-2]

            raise ValueError, msg
            
        if tmp_dimen is None:
            # We set the default to 1
            tmp_dimen = 1
        if self.initialize is not None:
            #
            # Convert generators to lists, since a generator is not copyable
            #
            if type(self.initialize) is types.GeneratorType:
                self.initialize = list(self.initialize)
            #
            # Try to guess dimen from the initialize list
            #
            tmp=None
            if type(self.initialize) is tuple:
                tmp = len(self.initialize)
            elif type(self.initialize) is list and len(self.initialize) > 0 \
                     and type(self.initialize[0]) is tuple:
                tmp = len(self.initialize[0])
            else:
                tmp = getattr(self.initialize, 'dimen', tmp)
            if tmp is not None:
                if kwd_dimen is not None and tmp != kwd_dimen:
                    raise ValueError, "Dimension argument differs from the data in the initialize list"
                else:
                    tmp_dimen = tmp

        self.dimen = tmp_dimen

    def reset(self):
        pass

    def dim(self):
        return self._ndim

    def display(self, ostream=None):
        self.pprint(ostream=ostream)

    def __getattr__(self,name):
        if name is "within":
           return self.domain
        raise AttributeError, "Unknown attribute "+name

    def __setattr__(self,name,val):
        if name is "within":
           self.domain = val
        return object.__setattr__(self,name,val)

    def _verify(self,element,use_exception=True):
        # A utility routine that is used to verify if the element
        # is valid for this set.
        if self.domain is not None and element not in self.domain:
           if use_exception:
              raise ValueError, "The value="+str(element)+" is not valid for set="+self.name+", because it is not within the domain="+self.domain.name
           return False
        if self.validate is not None and not self.validate(element,self.model):
           if use_exception:
              raise ValueError, "The value="+str(element)+" violates the validation rule of set="+self.name
           return False
        if self.dimen > 1 and type(element) is not tuple:
           if use_exception:
              raise ValueError, "The value="+str(element)+" is not a tuple for set="+self.name+", which has dimen="+str(self.dimen)
           return False
        elif self.dimen == 1 and type(element) is tuple:
           if use_exception:
              raise ValueError, "The value="+str(element)+" is a tuple for set="+self.name+", which has dimen="+str(self.dimen)
           return False
        elif type(element) is tuple and len(element) != self.dimen:
           if use_exception:
              raise ValueError, "The value="+str(element)+" does not have dimension="+str(self.dimen)+", which is needed for set="+self.name
           return False
        return True

    def bounds(self):
        """Return bounds information.  The default value is 'None', which
        indicates that this set does not contain bounds."""
        return self._bounds


class _SetContainer(_BaseSet):
    """A derived _BaseSet object that contains a single set."""

    def __init__(self, *args, **kwds):
        """ Constructor """
        if args != ():
            raise TypeError, "A _SetContainer expects no arguments"
        self._index_set=None
        self._index=[None]
        self._ndim=0
        self.value=set()
        _BaseSet.__init__(self,**kwds)

    def construct(self, values=None):
        """ Apply the rule to construct values in this set """
        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing _SetContainer, name="+self.name+", from data="+repr(values))
        if self._constructed:
            return
        self._constructed=True

        if self.initialize is None:
            self.initialize = getattr(self,'rule',None)
        #
        # Construct using the values list
        #
        if values is not None:
           if type(self._bounds) is tuple:
                first=self._bounds[0]
                last=self._bounds[1]
           else:
                first=None
                last=None
           all_numeric=True
           for val in values[None]:
                if type(val) in [int,float,long]:
                    if first is None or val<first:
                        first=val
                    if last is None or val>last:
                        last=val
                else:
                    all_numeric=False
                self.add(val)
           if all_numeric:
                self._bounds = (first,last)
        #
        # Construct using the rule
        #
        elif type(self.initialize) is types.FunctionType:
           if self.model is None:
              raise ValueError, "Must pass a model in to initialize with a function"
           if self.initialize.func_code.co_argcount == 1:
              #
              # Using a rule of the form f(model) -> iterator
              #
              tmp = self.initialize(self.model)
              for val in tmp:
                if self.dimen is None:
                   if type(val) in [tuple,list]:
                        self.dimen=len(val)
                   else:
                        self.dimen=1
                self.add(val)
           else:
              #
              # Using a rule of the form f(z,model) -> element
              #
              ctr=1
              val = self.initialize(ctr,self.model)
              if self.dimen is None:
                   if type(val) in [tuple,list]:
                        self.dimen=len(val)
                   else:
                        self.dimen=1
              while val is not None:
                self.add(val)
                ctr += 1
                val = self.initialize(ctr,self.model)
           #
           # We should be verifying as part of the add() method
           #
           #for val in self.value:
             #self._verify(val)
        #
        # Construct using the default values
        #
        elif self.initialize is not None:
              if type(self.initialize) is dict:
                 raise ValueError, "Cannot initialize set "+self.name+" with dictionary data"
              if type(self._bounds) is tuple:
                first=self._bounds[0]
                last=self._bounds[1]
              else:
                first=None
                last=None
              all_numeric=True
              for val in self.initialize:
                if type(val) in [int,float,long]:
                    if first is None or val<first:
                        first=val
                    if last is None or val>last:
                        last=val
                else:
                    all_numeric=False
                self.add(val)
              #print "HERE",self.name,first,last,all_numeric
              if all_numeric:
                self._bounds = (first,last)

        #
        # Apply filter, if defined
        #
        if self.filter is not None:
            self._apply_filter()

    def _apply_filter(self):
        """
        Applies self.filter to each element. If self.filter returns True,
        the element remain in the Set; otherwise, it is removed.
        """
        toRemove = []
        for el in self.value:
            if not self.filter(el, self.model):
                toRemove.append(el)

        for el in toRemove:
            self.remove(el)
            
    def data(self):
        """The underlying set data."""
        if self.virtual:
           raise TypeError, "Cannot access underlying set data for virtual set "+self.name
        return self.value

    def clear(self):
        """Remove all elements from the set."""
        if self.virtual:
           raise TypeError, "Cannot clear virtual Set object `"+self.name+"'"
        self.value.clear()
        if using_bidict:
            self.order_dict = bidict.bidict()
        else:
            self.order_dict = {}
            self.order_dict_inv = {}

    def check_values(self):
        """ Verify that the values in this set are valid.
        """
        if self.virtual:
           return
        for val in self.value:
          self._verify(val)

    def add(self, *args):
        """Add one or more elements to a set."""
        if self.virtual:
           raise TypeError, "Cannot add elements to virtual set `"+self.name+"'"
        for val in args:
          tmp = pyutilib.misc.flatten_tuple(val)
          self._verify(tmp)
          try:
            if tmp in self.value:
                continue
                #raise ValueError, "Element "+str(tmp)+" already exists in set "+self.name
            self.value.add(tmp)
            if self.ordered:
               if tmp in self.order_dict:
                    raise ValueError, "Element "+str(tmp)+" already exists in ordered set "+self.name
               self.order.append(tmp)
               if using_bidict:
                    self.order_dict[tmp:] = len(self.order)
               else:
                    self.order_dict[tmp] = len(self.order)
                    self.order_dict_inv[len(self.order)] = tmp
               

          except TypeError:
            raise TypeError, "Problem inserting "+str(tmp)+" into set "+self.name

    def remove(self, element):
        """Remove an element from the set.
           If the element is not a member, raise an error.
        """
        if self.virtual:
            raise KeyError, "Cannot remove element `"+str(element)+"' from virtual set "+str(self.name)
        if element not in self.value:
            raise KeyError, "Cannot remove element `"+str(element)+"' from set "+str(self.name)
        self.value.remove(element)
        if self.ordered:
            id = self.order_dict[element]-1
            for i in xrange(id+1,len(self.order)):
                if using_bidict:
                    self.order_dict[self.order[i]] = i
                else:
                    self.order_dict[self.order[i]] = i
                    self.order_dict_inv[i] = self.order[i]
            del self.order[id]

    def discard(self, element):
        """Remove an element from the set.
           If the element is not a member, do nothing.
        """
        if self.virtual:
           raise KeyError, "Cannot discard element `"+str(element)+"' from virtual set "+str(self.name)
        if self.ordered and element in self.value:
           self.order.remove(element)
        self.value.discard(element)

    def first(self):
        if self.virtual:
            raise TypeError, "Cannot access the first element of virtual set `"+self.name+"'"
        return self.member(1)

    def last(self):
        if self.virtual:
            raise TypeError, "Cannot access the last element of virtual set `"+self.name+"'"
        return self.member(len(self.value))

    def __getitem__(self, key):
        return self.member(key)

    def member(self, key):
        if self.ordered:
            if using_bidict:
                return self.order_dict.inv[key]
            else:
                return self.order_dict_inv[key]
        #
        # If the set is not ordered, then we use the intrinsic ordering imposed by the set.
        # We convert the set to a list and return the specified value.
        #
        if key >= 1:
            return list(self.value)[key-1]
        elif key < 0:
            return list(self.value)[key]
        else:
            raise IndexError, "Valid index values for sets are 1 ... len(set) or -1 ... -len(set)"

    def ord(self, match_element):
        """ For ordered sets, return the position index (1-based) 
            of the input element.
        """
        if self.ordered is False:
           raise AttributeError, "Cannot invoke ord() method for unordered set="+str(self.name)
        try:
            return self.order_dict[match_element]
        except IndexError:
            raise IndexError, "Unknown input element="+str(match_element)+" provided as input to ord() method for set="+str(self.name)

    def next(self, match_element, k=1):
        # NOTE: we currently aren't consist (and don't actually deal with) w.r.t. unordered sets - see the implementation of member() above.
        try:
           element_position = self.order_dict[match_element]
        except KeyError:
           raise KeyError, "Cannot obtain next() member of set="+self.name+"; input element="+str(match_element)+" is not a member of the set!"

        try:
           return self.order[element_position+k-1] # "order" is a list, accessed 0-based.
        except KeyError:
           raise KeyError, "Cannot obtain next() member of set="+self.name+"; failed to access item in position="+str(element_position+k-1)

    def nextw(self, match_element, k=1):
        ndx = self.order_dict[match_element]+k
        total = len(self.order)
        if ndx > total:
            ndx -= total
        if ndx <= 0:
            ndx += total
        return self.order.inv[ndx]

    def prev(self, match_element, k=1):
        return self.next(match_element, k=-k)

    def prevw(self, match_element, k=1):
        return self.nextw(match_element, k=-k)

    def pprint(self, ostream=None):
        if ostream is None:
           ostream = sys.stdout
        print >>ostream, "  ",self.name,":",
        print >>ostream, "\tDim="+str(self.dim()),
        print >>ostream, "\tDimen="+str(self.dimen),
        print >>ostream, "\tSize="+str(len(self)),
        if self.domain is not None:
           print >>ostream, "\tDomain="+str(self.domain.name),
        else:
           print >>ostream, "\tDomain="+str(None),
        print >>ostream, "\tOrdered="+str(self.ordered),
        print >>ostream, "\tBounds="+str(self._bounds)
        if self.model is None:
           print >>ostream, "\t Model=None"
        else:
           print >>ostream, "\t Model="+str(self.model.name)
        if self.virtual:
           print >>ostream, "\t  Virtual"
        else:
           if self.ordered:
              tmp = copy.copy(self.order)
           else:
              tmp = copy.copy(list(self.value))
              tmp.sort()
           print >>ostream, "\t  ",tmp

    def __len__(self):
        """ The number of items in the set.
            The set is empty until a concrete instance has been setup.
        """
        if self.virtual:
           raise ValueError, "The size of a virtual set is unknown"
        return len(self.value)

    #def __repr__(self):
        #"""Return string representation of the underlying set"""
        #return str(self)+" : "+str(self.value)

    def __iter__(self):
        """Return an iterator for the underlying set"""
        if self.virtual:
           raise TypeError, "Cannot iterate over Set object `"+self.name+"', which is a virtual set"
        if self.ordered:
           return self.order.__iter__()
        return self.value.__iter__()

    def __hash__(self):
        """Hash this object"""
        return _BaseSet.__hash__(self)

    def __eq__(self,other):
        """ Equality comparison """
        try:
            tmp = self._set_repn(other)
        except TypeError:
            # An exception is raised if a non-set object is passed to _set_repn
            return False
        if self.virtual:
            if other.virtual:
                return hash(self) == hash(tmp)
            return False
        if self.dimen != other.dimen:
           return False
        return self.value.__eq__( tmp )

    def __ne__(self,other):
        return not self.__eq__(other)

    def __contains__(self, element):
        """Report whether an element is a member of this set.
        (Called in response to the expression 'element in self'.)
        """
        #
        # If the element is a set, then see if this is a subset
        #
        if isinstance(element,_SetContainer):
           return element.issubset(self)
        #
        # If this is not a valid element, then return False
        #
        if not self._verify(element,False):
           return False
        #
        # If the restriction rule is used then we do not actually
        # check whether the data is in the set self.value.
        #
        if self.validate is not None and self.virtual:
           return True
        #
        # The final check: return true if self.virtual is True, since we should
        # have already validated this value
        #
        return self.virtual or element in self.value

    def issubset(self,other):
        """Report whether another set contains this set"""
        if self.virtual:
           raise TypeError, "ERROR: cannot perform \"issubset\" test because the current set is a virtual set."
        if self.dimen != other.dimen:
           raise ValueError, "Cannot perform set operation with sets "+self.name+" and "+other.name+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen)
        for val in self.value:
          if val not in other:
             return False
        return True

    def issuperset(self,other):
        """Report this set contains another set"""
        if isinstance(other,_BaseSet):
           if other.virtual:
              raise TypeError, "ERROR: cannot perform \"issuperset\" test because the target set is a virtual set."
           return other.issubset(self) and (self.virtual or len(self) >= len(other))
        for val in other:
          if val not in self:
             return False
        return True

    __le__ = issubset
    __ge__ = issuperset

    def __lt__(self,other):
        if self.virtual or \
           (isinstance(other,_BaseSet) and other.virtual):
           raise TypeError, "ERROR: cannot test for __lt__ because either or both of the sets are virtual."
        if self.dimen != other.dimen:
           raise ValueError, "Cannot perform set operation with sets "+self.name+" and "+other.name+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen)
        for val in self.value:
          if val not in other:
             return False
        return len(self) < len(other)

    def __gt__(self,other):
        if self.virtual or \
           (isinstance(other,_BaseSet) and other.virtual):
           raise TypeError, "ERROR: cannot test for __lt__ because either or both of the sets are virtual."
        if self.dimen != other.dimen:
           raise ValueError, "Cannot perform set operation with sets "+self.name+" and "+other.name+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen)
        for val in other:
          if val not in self.value:
             return False
        return len(self) > len(other)

    def _set_repn(self,other):
        if isinstance(other,_SetContainer) and other.virtual:
            return other
        if isinstance(other,_SetContainer) and type(other.value) is not dict:
           return other.value
        raise TypeError, "Cannot apply Set operation with a "+str(type(other))+" object"
        
    def __or__(self,other):
        """ Return a _SetContainer object that is the union of this set
        with another set. """
        if self.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+self.name+"'"
        if other.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+other.name+"'"
        if self.dimen != other.dimen:
           raise ValueError, "Cannot perform set operation with sets "+self.name+" and "+other.name+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen)
        def SetUnionRule(model):
            data = set()
            for val in getattr(model,SetUnionRule.Set1):
              data.add(val)
            for val in getattr(model,SetUnionRule.Set2):
              data.add(val)
            return data
        SetUnionRule.Set1 = self.name
        SetUnionRule.Set2 = other.name
        return _SetContainer(dimen=self.dimen, initialize=SetUnionRule)

    def __and__(self,other):
        """Return a _SetContainer object that is the intersection of this set
        with another set"""
        if self.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+self.name+"'"
        if other.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+other.name+"'"
        if self.dimen != other.dimen:
           raise ValueError, "Cannot perform set operation with sets "+self.name+" and "+other.name+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen)

        if (self._constructed is True) and (other._constructed is True):
           data = []
           for val in self:
              if val in other:
                 data.append(val)
           result = _SetContainer(dimen=self.dimen, initialize=data)
           # IMPT: construct here, as the & operator is often used in-line, such that the component is not part of any model - and therefore won't be constructed when it is added to a model!
           result.construct()
           return result
        else:
           def SetIntersectionRule(model):
               data = set()
               for val in getattr(model,SetIntersectionRule.Set1):
                 if val in getattr(model,SetIntersectionRule.Set2):
                    data.add(val)
               return data
           SetIntersectionRule.Set1 = self.name
           SetIntersectionRule.Set2 = other.name
           return _SetContainer(dimen=self.dimen, initialize=SetIntersectionRule)

    def __xor__(self,other):
        """Return a _SetContainer object that is the symmetric difference of 
        this set with another set"""
        if self.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+self.name+"'"
        if other.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+other.name+"'"
        if self.dimen != other.dimen:
           raise ValueError, "Cannot perform set operation with sets "+self.name+" and "+other.name+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen)
        def SetSymmetricDifferenceRule(M):
            Set1 = getattr(M,SetSymmetricDifferenceRule.Set1)
            Set2 = getattr(M,SetSymmetricDifferenceRule.Set2)
            data = set()
            for val in Set1:
              if val not in Set2:
                 data.add(val)
            for val in Set2:
              if val not in Set1:
                 data.add(val)
            return data
        SetSymmetricDifferenceRule.Set1 = self.name
        SetSymmetricDifferenceRule.Set2 = other.name
        return _SetContainer(dimen=self.dimen, initialize=SetSymmetricDifferenceRule)

    def __sub__(self,other):
        """Return a _SetContainer object that is the difference between this set
        and another set"""
        if self.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+self.name+"'"
        if other.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+other.name+"'"
        if self.dimen != other.dimen:
           raise ValueError, "Cannot perform set operation with sets "+self.name+" and "+other.name+" that have different element dimensions: "+str(self.dimen)+" "+str(other.dimen)
        def SetDifferenceRule(model):
            data = set()
            for val in getattr(model,SetDifferenceRule.Set1):
              if val not in getattr(model,SetDifferenceRule.Set2):
                 data.add(val)
            return data
        SetDifferenceRule.Set1 = self.name
        SetDifferenceRule.Set2 = other.name
        return _SetContainer(dimen=self.dimen, initialize=SetDifferenceRule)

    def __mul__(self,other):
        """Return a _ProductSet that is the cross-product between this set and
        and another set"""
        if not isinstance(other,_SetContainer):
           raise TypeError, "Set operator __mul__ cannot be applied with an object of type "+str(type(other))
        if self.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+self.name+"'"
        if other.virtual:
           raise TypeError, "Cannot perform set operations with virtual set '"+other.name+"'"
        if isinstance(self,_ProductSet) and self.name=="_unknown_":
           self.set_tuple = tuple( list(self.set_tuple) + [other] )
           self._compute_dimen()
           return self
        return _ProductSet(self,other)


class _SetArray(_BaseSet):
    """An array of sets, which are indexed by other sets"""

    def __init__(self, *args, **kwds):      #pragma:nocover
        """ Constructor """
        self._index_set=None
        self.value={}
        self._ndim=len(args)
        if len(args) == 1:
            if isinstance(args[0],_BaseSet):
                self._index=args[0]
            else:
                try:
                    options = getattr(args[0],'set_options')
                    options['initialize'] = args[0]
                    self._index=Set(**options)
                except:
                    self._index=Set(initialize=args[0])
        else:
            self._index=None
            tmp = []
            for arg in args:
                if isinstance(arg,_BaseSet):
                    tmp.append(arg)
                else:
                    try:
                        options = getattr(arg,'set_options')
                        options['initialize'] = arg
                        tmp.append( Set(**options) )
                    except:
                        tmp.append( Set(initialize=arg) )
            self._index_set=tuple(tmp)
        _BaseSet.__init__(self,**kwds)
        self.value={}

    def keys(self):
        return self.value.keys()

    def __contains__(self, element):
        return element in self.value.keys()

    def clear(self):
        """Remove all elements from the set."""
        for key in self.value:
          self.value[key].clear()

    def data(self):
        """The underlying set data."""
        raise TypeError, "Cannot access underlying set data for array set "+self.name

    def __getitem__(self, key):
        if key not in self.value:
           if (key in self._index) and (self.initialize is not None) and (type(self.initialize) in [tuple,list]):
              return self.initialize
           raise KeyError, "Cannot access index "+str(key)+" in array set "+self.name
        return self.value[key]

    def __setitem__(self, key, val):
        if key not in self._index:
           raise KeyError, "Cannot set index "+str(key)+" in array set "+self.name
        if key in self.value:
           self.value[key].clear()
           for elt in val:
             self.value[key].add(elt)
        else:
           self.value[key] = Set(initialize=val,within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(key)+"]",dimen=self.dimen)
           self.value[key].construct()

    def __len__(self):
        result = 0
        for val in self.value.values():
            result += len(val)
        return result

    def __iter__(self):
        """Return an iterator for the set array"""
        return self.value.__iter__()

    def check_values(self):
        for key in self.value:
          for val in self.value[key]:
            self._verify(val,True)

    def pprint(self, ostream=None):
        if ostream is None:
           ostream = sys.stdout
        print >>ostream, "  ",self.name,":",
        print >>ostream, "\tDim="+str(self.dim()),
        print >>ostream, "\tDimen="+str(self.dimen),
        if self.domain is None:
           print >>ostream, "\tDomain="+str(self.domain),
        else:
           print >>ostream, "\tDomain="+str(self.domain.name),
        print >>ostream, "\tArraySize="+str(len(self.value.keys())),
        print >>ostream, "\tOrdered="+str(self.ordered)
        if self.model is None:
           print >>ostream, "\t Model=None"
        else:
           print >>ostream, "\t Model="+str(self.model.name)        
        tmp=self.value.keys()
        if tmp is not None:
            tmp.sort()
        else:
            tmp=[]
        for val in tmp:
          if self.ordered:
             ttmp = copy.copy(self.value[val].order)
          else:
             ttmp = copy.copy(list(self.value[val].value))
             ttmp.sort()
          print >>ostream, "\t",self.name+"["+str(val)+"]\t",ttmp

    def construct(self, values=None):
        """ Apply the rule to construct values in each set"""
        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing _SetArray, name="+self.name+", from data="+repr(values))
        if self._constructed:
            return
        self._constructed=True
        if self.virtual:                                #pragma:nocover
           raise TypeError, "It doesn't make sense to create a virtual set array"

        if self.initialize is None:
            self.initialize = getattr(self,'rule',None)

        #
        # Construct using the values list
        #
        if values is not None:
           for key in values:
             if type(key) is tuple and len(key)==1:
                tmpkey=key[0]
             else:
                tmpkey=key
             if tmpkey not in self._index:
                raise KeyError, "Cannot set index "+str(tmpkey)+" in array set "+self.name
             self.value[tmpkey] = Set(initialize=values[key],within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(tmpkey)+"]",dimen=self.dimen)
             self.value[tmpkey].model = self.model
             self.value[tmpkey].construct()
        #
        # Construct using the rule
        #
        elif type(self.initialize) is types.FunctionType:
           if self.model is None:
              raise ValueError, "Need model to construct a set array with a function"
           if self._index is None:
              raise ValueError, "No index for set "+self.name
           for key in self._index:
             if isinstance(key,tuple):
                tmp = list(key)
             else:
                tmp = [key]
             if self.initialize.func_code.co_argcount == len(tmp)+1:
                tmp.append(self.model)
                tmp = tuple(tmp)
                rule_list = list(self.initialize(*tmp))
             else:
                rule_list=[]
                ctr=1
                args = tuple(tmp+[ctr,self.model])
                val = self.initialize(*args)
                while val is not None:
                  rule_list.append(val)
                  ctr += 1
                  args = tuple(tmp+[ctr,self.model])
                  val = self.initialize(*args)
             if self.dimen is None and len(rule_list) > 0:
                   if type(rule_list[0]) in [tuple,list]:
                        self.dimen=len(rule_list[0])
                   else:
                        self.dimen=1
             self.value[key] = Set(initialize=rule_list,within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(key)+"]",dimen=self.dimen)
             self.value[key].model = self.model
             self.value[key].construct()
        #
        # Construct using the default values
        #
        else:
           if self.initialize is not None:
              if type(self.initialize) is not dict:
                 for key in self._index:
                   self.value[key] = Set(initialize=self.initialize,within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(key)+"]",dimen=self.dimen)
                   self.value[key].model = self.model
                   self.value[key].construct()
              else:
                 for key in self.initialize:
                   self.value[key] = Set(initialize=self.initialize[key],within=self.domain,validate=self.validate,ordered=self.ordered,name=self.name+"["+str(key)+"]",dimen=self.dimen)
                   self.value[key].model = self.model
                   self.value[key].construct()


##------------------------------------------------------------------------
##
## A set that defines an abstract cross-product set.
##
##------------------------------------------------------------------------

class _ProductSet(_SetContainer):
    """Set that represents a cross-product of other sets"""

    """ Constructor
        Arguments:
           name         The name of this set
           initialize   Default set members, which may be overriden
                            when setting up this set
           rule         A rule for setting up this set with
                            existing model data.  This has the functional form:
                            f: pyomo.Model -> set
    """
    def __init__(self, *args, **kwd):
        self._class_override=False

        # a base set container doesn't allow arguments - 
        # avoiding weird (and possibly incorrect) *()
        # syntax below.
        tmp=()
        _SetContainer.__init__(self,*tmp,**kwd)

        # the individual index sets definining the product set. 
        self.set_tuple = args

        # sets the "dimen" instance attribute.
        self._compute_dimen()

        # _len is the number of elements in the product 
        # set, computed in the obvious way. 
        # NOTE: this is not dynamically updated, so if
        #       the composite set contents are updated,
        #       the change will not be reflected here.
        self._len=0

    def __iter__(self):
        """Returns an iterator/generator for the cross-product""" 

        if sys.version_info[:2] >= (2,6):
            if self.is_flat_product() is True:
                return (i for i in itertools.product(*self.set_tuple))               
            else:
                return (pyutilib.misc.flatten_tuple(i) for i in itertools.product(*self.set_tuple))
        else:
            return pyutilib.misc.flattened_cross_iter(*self.set_tuple)

    def __len__(self):
        return self._len

    def _compute_dimen(self):
        ans=0
        for set in self.set_tuple:
            if set.dimen is None:
                self.dimen=None
                return
            else:
                ans += set.dimen
        self.dimen = ans

    def is_flat_product(self):
        """
        a simple utility to determine if each of the composite sets is 
        of dimension one. knowing this can significantly reduce the 
        cost of iteration, as you don't have to call flatten_tuple.
        """

        for s in self.set_tuple:
           if s.dimen != 1:
              return False
        return True

    def _verify(self,element,use_exception=True):
        """
        If this set is virtual, then an additional check is made
        to ensure that the element is in each of the underlying sets.
        """
        tmp = _SetContainer._verify(self,element,use_exception)
        if not tmp or not self.virtual:
            return tmp

        next_tuple_index = 0
        member_set_index = 0
        for member_set in self.set_tuple:
           tuple_slice = element[next_tuple_index:next_tuple_index + member_set.dimen]
           if member_set.dimen == 1:
              tuple_slice = tuple_slice[0]
           if tuple_slice not in member_set:
             return False
           member_set_index += 1
           next_tuple_index += member_set.dimen
        return True
        
    def construct(self, values=None):
        """ Apply the rule to construct values in this set """
        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing _ProductSet, name="+self.name+", from data="+repr(values))
        if self._constructed:
            return
        self._constructed=True

        if self.initialize is None:
            self.initialize = getattr(self,'rule',None)

        if self.virtual:
           if len(self.set_tuple) == 0:
              self._len=0
           else:
              self._len=1
              for val in self.set_tuple:
                self._len *= len(val)
        #
        # Construct using the values list
        #
        elif values is not None:
           for val in values[None]:
             self.add(val)
        #
        # Construct using the rule
        #
        elif type(self.initialize) is types.FunctionType:
           for val in self.initialize(self.model):
             self.add(val)
        #
        # Construct using the default values
        #
        else:
           if self.initialize is not None:
              if type(self.initialize) is dict:
                 for val in self.initialize:
                   self.add(val)
              else:
                 self.add(self.initialize)
           else:
              for val in pyutilib.misc.cross(self.set_tuple):
                self.add(val)
        #
        if not self.virtual:
            self._len=len(self.value)
        self._compute_dimen()
        if self.dimen is None:
            raise ValueError, "The product set dimension must be computeable after constructing this set!"


class Set(Component):
    """Set objects that are used to index other Pyomo objects

       This class has a similar look-and-feel as a built-in set class.  However,
       the set operations defined in this class return another abstract
       Set object.  This class contains a concrete set, which can be 
       initialized by the load() method.
    """

    alias('Set', "Set data that is used to define a model instance.")

    def __new__(cls, *args, **kwds):
        if args == ():
            self = _SetContainer(*args, **kwds)
        else:
            self = _SetArray(*args, **kwds)
        return self




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________

from __future__ import division

__all__ = ['generate_canonical_repn', 'as_expr', 'is_constant', 'is_linear', 'is_quadratic', 'is_nonlinear']

import logging
from coopr.pyomo.base import IPyomoPresolver, IPyomoPresolveAction, Model, \
                             Constraint, Objective
from coopr.pyomo.base import expr
from coopr.pyomo.base.var import _VarValue, Var
import copy


logger = logging.getLogger('coopr.pyomo')
#
# A frozen dictionary that can be hashed.  This dictionary isn't _really_
# frozen, but it acts hashable.
#
class frozendict(dict):

    __slots__ = ['_hash']

    def __hash__(self):
        rval = getattr(self, '_hash', None)
        if rval is None:
            rval = self._hash = hash(frozenset(self.iteritems()))
        return rval

    # getstate and setstate are currently not defined in a generic manner, i.e.,
    # we assume the underlying dictionary is the only thing 

    def __getstate__(self):
       return self.items()

    def __setstate__(self, dictionary):
       self.clear()
       self.update(dictionary)
       self._hash = None

#
# Generate a canonical representation of an expression.
#
# The canonical representation is a dictionary.  Each element is a mapping
# from a term degree to terms in the expression.  If the term degree is
# None, then the map value is simply an expression of terms without a
# specific degree.  Otherwise, the map value is a dictionary of terms to
# their coefficients.  A term is represented as a frozen dictionary that
# maps variable id to variable power.  A constant term is represented
# with None.
#
# Examples:
#  Let x[1] ... x[4] be the first 4 variables, and
#      y[1] ... y[4] be the next 4 variables
#
# 1.3                           {0:{ None :1.3}}
# 3.2*x[1]                      {1:{ {0:1} :3.2}}
# 2*x[1]*y[2] + 3*x[2] + 4      {0:{None:4.0}, 1:{{1:1}:3.0}, 2:{{0:1, 5:1}:2.0}}
# 2*x[1]**4*y[2]    + 4         {0:{None:4.0}, 1:{{1:1}:3.0}, 5:{{0:4, 5:1 }:2.0}}
# log(y[1]) + x[1]*x[1]         {2:{{0:2}:1.0}, None:log(y[1])}
#

def generate_canonical_repn(exp):

    degree = exp.polynomial_degree()
    if degree == 1:
        # varmap is a map from the variable hash to a VarValue.
        # coef is a map from the variable hash to it's coefficient.
        coef, varmap = collect_linear_canonical_repn(exp)
        ans = {}
        if None in coef:
            val = coef.pop(None)
            if val != 0:
                ans[0] = { None: val }
        tmp = ans.setdefault(1,{})
        for key, val in coef.iteritems():
            tmp[frozendict({key:1})] = val
        ans[-1] = varmap
        return ans
    elif degree > 1:
        return collect_general_canonical_repn(exp)
    elif degree == 0:
        return { 0: { None:exp() } }
    else:
        return { None: exp, -1 : collect_variables(exp) }


def as_expr(rep, vars=None, model=None, ignore_other=False):
    """ Convert a canonical representation into an expression. """
    if isinstance(model, Model):
        vars = model._var
        id_offset=0
    elif not vars is None:
        id_offset=1
    exp = 0.0
    for d in rep:
        if d is None:
            if not ignore_other:
                exp += rep[d]
            continue
        if d is -1:
            continue
        for v in rep[d]:
            if v is None:
                exp += rep[d][v]
                continue
            e = rep[d][v]
            for id in v:
                for i in xrange(v[id]):
                    if vars is None:
                        e *= rep[-1][id]
                    else:
                        e *= vars[id[1]+id_offset]
            exp += e
    return exp
    
def repn_add(lhs, rhs, coef=1.0):
    """
    lhs and rhs are the expressions being added together.
    'lhs' and 'rhs' are left-hand and right-hand side operands
    See generate_canonical_repn for explanation of pyomo expressions
    """
    for order in rhs:
        # For each order term, first-order might be 3*x,
        # second order 4*x*y or 5*x**2
        if order is None:
            # i.e., (currently) order is a constant or logarithm
            if order in lhs:
                lhs[order] += rhs[order]
            else:
                lhs[order] = rhs[order]
            continue
        if order < 0:
            # ignore now, handled below
            continue
        if not order in lhs:
            lhs[order] = {}
        for var in rhs[order]:
            # Add coefficients of variables in this order (e.g., third power)
            lhs[order][var] = coef*rhs[order][var] + lhs[order].get(var,0.0)
    #
    # Merge the VarValue maps
    #
    if -1 in rhs:
        if -1 in lhs:
            lhs[-1].update(rhs[-1])
        else:
            lhs[-1] = rhs[-1]
    return lhs


def repn_mult(r1, r2, coef=1.0):
    rep = {}
    for d1 in r1:
        for d2 in r2:
            if d1 == None or d2 == None or d1 < 0 or d2 < 0:
                pass
            else:
                d=d1+d2
                if not d in rep:
                    rep[d] = {}
                if d == 0:
                        rep[d][None] = coef * r1[0][None] * r2[0][None]
                elif d1 == 0:
                        for v2 in r2[d2]:
                            rep[d][v2]  = coef * r1[0][None] * r2[d2][v2] + rep[d].get(v2,0.0)
                elif d2 == 0:
                        for v1 in r1[d1]:
                            rep[d][v1]  = coef * r1[d1][v1] * r2[0][None] + rep[d].get(v1,0.0)
                else:
                        for v1 in r1[d1]:
                            for v2 in r2[d2]:
                                v = frozendict(v1)
                                for id in v2:
                                    if id in v:
                                        v[id] += v2[id]
                                    else:
                                        v[id]  = v2[id]
                                rep[d][v] = coef * r1[d1][v1] * r2[d2][v2] + rep[d].get(v,0.0)
    #
    # Handle other nonlinear terms
    #
    if None in r1:
        rep[None] = as_expr(r2, ignore_other=True) * copy.deepcopy(r1[None])
        if None in r2:
            rep[None] += copy.deepcopy(r1[None])*copy.deepcopy(r2[None])
    if None in r2:
        if None in rep:
            rep[None] += as_expr(r1, ignore_other=True) * copy.deepcopy(r2[None])
        else:
            rep[None]  = as_expr(r1, ignore_other=True) * copy.deepcopy(r2[None])
    #
    # Merge the VarValue maps
    #
    if -1 in r1:
        rep[-1] = r1[-1]
    if -1 in r2:
        if -1 in rep:
            rep[-1].update(r2[-1])
        else:
            rep[-1] = r2[-1]
    #
    # Return the canonical repn
    #
    return rep


def collect_variables(exp):
    if exp.is_expression():
        ans = {}
        if exp.__class__ is expr._ProductExpression:
            for subexp in exp._numerator:
                ans.update(collect_variables(subexp))
            for subexp in exp._denominator:
                ans.update(collect_variables(subexp))
        else:
            # This is fragile: we assume that all other expression
            # objects "play nice" and just use the _args member.
            for subexp in exp._args:
                ans.update(collect_variables(subexp))
        return ans
    elif exp.is_constant():
        # NB: is_constant() returns True for variables with fixed values
        return {}
    elif exp.__class__ is _VarValue:
        key = ( id(exp.var.model), exp.id )
        return { key : exp }
    elif exp.type() is Var:
        key = ( id(exp.model), exp.id )
        return { key : exp }
    else:
       raise ValueError, "Unexpected expression type: "+str(exp)

def collect_linear_canonical_repn(exp):
    #
    # Ignore identity expressions
    #

    while exp.__class__ is expr._IdentityExpression:
        exp = exp._args[0]

    # constants and _VarValues are more frequent than expressions - look for them first, for efficiency.
    if exp.is_constant():
        # NB: is_constant() returns True for variables with fixed values
        return { None: exp() }, {}
    elif exp.__class__ is _VarValue:
        key = ( id(exp.var.model), exp.id )
        return { key : 1.0 }, { key : exp }
    elif exp.is_expression():
        if exp.__class__ is expr._ProductExpression:
            # OK - the following is safe ONLY because we verify the
            # expression order before entering here -- so we know that
            # there will be no more than 1 variable in the product
            # expression (and that that variable will appear in the
            # numerator!)
            coef = None
            varmap = None
            const = exp.coef
            for subexp in exp._denominator: 
                x = subexp() # only have constants in the denominator.
                if x == 0:
                    import StringIO
                    buf = StringIO.StringIO()
                    subexp.pprint(buf)
                    logger.error("Divide-by-zero: offending sub-expression:\n   " + buf)
                    raise ZeroDivisionError                        
                const /= x
            for subexp in exp._numerator:
                _coef, _varmap = collect_linear_canonical_repn(subexp)
                if _varmap:
                    varmap = _varmap
                    coef = _coef
                else:
                    const *= _coef[None]

            if coef:
                if const != 1:
                    for var in coef.iterkeys():
                        coef[var] *= const
                return coef, varmap
            else:
                return { None: const }, {}
        elif exp.__class__ is expr._SumExpression:
            coef = { None : exp._const } # None is the constant term in the coefficient map.
            varmap = {}
            for i in xrange(len(exp._args)):

                arg_coef = exp._coef[i]
                arg = exp._args[i]
                
                if (arg.__class__ is _VarValue) and (arg.fixed is False): 
                   # save an expensive recursion - this is by far the most common case.
                   var_key = ( id(arg.var.model), arg.id )
                   varmap.setdefault(var_key, arg)
                   coef[var_key] = coef.get(var_key, 0.0) + arg_coef
                else:
                   _coef, _varmap = collect_linear_canonical_repn(arg)
                   varmap.update(_varmap)
                   for var, val in _coef.iteritems():
                       coef[var] = coef.get(var,0.0) + val * arg_coef

            return coef, varmap
        else:
            raise ValueError, "Unsupported expression type: "+str(exp)
    elif exp.type() is Var:
        key = ( id(exp.model), exp.id )
        return { key : 1.0 }, { key : exp }
    else:
       raise ValueError, "Unexpected expression type: "+str(exp)
        

#
# Temporary canonical expressions
#
# temp_const = { 0: {None:0.0} }    
# temp_var = { 1: {frozendict({None:1}):1.0} }    
# temp_nonl = { None: None }

#
# Internal function for collecting canonical representation, which is
# called recursively.
#
def collect_general_canonical_repn(exp):
#     global temp_const
#     global temp_var
#     global temp_nonl
    temp_const = { 0: {None:0.0} }    
    temp_var = { 1: {frozendict({None:1}):1.0} }    
    temp_nonl = { None: None }    
    #
    # Ignore identity expressions
    #
    while isinstance(exp, expr._IdentityExpression):
        exp = exp._args[0]
    #
    # Expression
    #
    if isinstance(exp,expr.Expression):
        #
        # Sum
        #
        if isinstance(exp,expr._SumExpression):
            if exp._const != 0.0:
                repn = { 0: {None:exp._const} }    
            else:
                repn = {}
            for i in xrange(len(exp._args)):
                repn = repn_add(repn, collect_general_canonical_repn(exp._args[i]), coef=exp._coef[i] )
            return repn
        #
        # Product
        #
        elif isinstance(exp,expr._ProductExpression):
            #
            # Iterate through the denominator.  If they aren't all constants, then
            # simply return this expresion.
            #
            denom=1.0
            for e in exp._denominator:
                if e.fixed_value():
                    #print 'Z',type(e),e.fixed
                    denom *= e.value
                elif e.is_constant():
                    denom *= e()
                else:
                    temp_nonl[None] = exp
                    return temp_nonl
                if denom == 0.0:
                    print "Divide-by-zero error - offending sub-expression:"
                    e.pprint()
                    raise ZeroDivisionError                        
            #
            # OK, the denominator is a constant.
            #
            repn = { 0: {None:exp.coef / denom} }    
            #print "Y",exp.coef/denom
            for e in exp._numerator:
                repn = repn_mult(repn, collect_general_canonical_repn(e))
            return repn
        #
        # ERROR
        #
        else:
            raise ValueError, "Unsupported expression type: "+str(exp)
    #
    # Constant
    #
    elif exp.fixed_value():
        temp_const[0][None] = exp.value
        return temp_const
    #
    # Variable
    #
    elif type(exp) is _VarValue or exp.type() is Var:
        if type(exp) is _VarValue:
            mid = id(exp.var.model)
        else:
            mid = id(exp.model)
        temp_var = { -1: {(mid,exp.id):exp}, 1: {frozendict({(mid,exp.id):1}):1.0} }    
        return temp_var
    #
    # ERROR
    #
    else:
       raise ValueError, "Unexpected expression type: "+str(exp)



def is_constant(repn):
    """Return True if the canonical representation is a constant expression"""
    for key in repn:
        if key is None or key > 0:
            return False
    return True

def is_nonlinear(repn):
    """Return True if the canonical representation is a nonlinear expression"""
    for key in repn:
        if not key in [-1,0,1]:
            return True
    return False

def is_linear(repn):
    """Return True if the canonical representation is a linear expression."""
    return (1 in repn) and not is_nonlinear(repn)

def is_quadratic(repn):
    """Return True if the canonical representation is a quadratic expression."""
    if not 2 in repn:
        return False
    for key in repn:
        if key is None or key > 2:
            return False
    return True

